"""Exposes shared utilities and base classes for pydantic-redis

This includes basic functionality of mutating and querying
redis via the pydantic-redis ORM regardless of whether this
is done asynchronously or synchronously.
This is a private package.
"""
